#!/usr/bin/env python
"""Utility functions and classes for python"""
########################################################################
# File: utils.py
#  executable: utils.py
#
# Author: Andrew Bailey
# History: 12/09/17 Created
########################################################################

import logging
import os
import string
import random
import re
import json
import tarfile
import glob
from datetime import datetime
from contextlib import contextmanager
from io import StringIO
import sys
import numpy as np
import datetime
from collections import Iterable
from itertools import takewhile, repeat, islice


@contextmanager
def captured_output():
    new_out, new_err = StringIO(), StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err


def create_logger(file_path, name="a", info=False, debug=False):
    """Create a logger instance which will write all logs to file path and display logs at requested level
    :param file_path: path to file without .log extension
    :param name: unique name of logger
    :param info: set logging level to INFO
    :param debug: set logging level to DEBUG
    :return: logger which will write all log calls to file and display at level requested
    """
    assert type(info) is bool, "info should be bool, you passed {}".format(type(info))
    assert type(debug) is bool, "debug should be bool, you passed {}".format(type(debug))
    assert type(file_path) is str, "file path needs to be string"
    assert type(name) is str, "name needs to be string"

    level = logging.WARNING
    if debug:
        level = logging.DEBUG
    if info:
        level = logging.INFO
    # create name for logger
    root_logger = logging.getLogger(name)
    root_logger.setLevel(logging.DEBUG)
    # format log info formatting
    log_formatter = logging.Formatter("%(asctime)s [%(threadName)s] [%(levelname)s]  %(message)s")

    file_handler = logging.FileHandler("{}.log".format(file_path))
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(log_formatter)

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(log_formatter)
    console_handler.setLevel(level)

    root_logger.addHandler(console_handler)
    root_logger.addHandler(file_handler)

    return logging.getLogger(name)


def create_dot_dict(dictionary):
    """Creates a dot.dictionary object and searches for internal dictionaries and turns them into dot.dictionaries

    note: only searches dictionaries and lists. sets or other iterable objects are not changed

    :param dictionary: a dictionary
    """
    # initialize dot.dict
    dictionary = DotDict(dictionary)
    # check if we need to change values within dictionary
    for key, value in dictionary.items():
        if isinstance(value, dict):
            dictionary[key] = create_dot_dict(value)
        if isinstance(value, list):
            new_list = []
            for x in value:
                if isinstance(x, dict):
                    new_list.append(create_dot_dict(x))
                else:
                    new_list.append(x)
            dictionary[key] = new_list

    return dictionary


class DotDict(dict):
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __init__(self, dictionary):
        """dot.notation access to dictionary attributes for string keys ONLY!

        :param dictionary: a dictionary"""
        super(DotDict, self).__init__(dictionary)
        for key, value in dictionary.items():
            assert type(key) is str, "Key must be a string"


def list_dir(path, ext="", include_dirs=False):
    """get all file paths from local directory with extension
    :param ext: optional extension for file
    :param include_dirs: boolean option to include directories
    :param path: path to directory
    :return: list of paths to files
    """
    assert type(path) is str, "Path must be a string"
    assert os.path.isdir(path), "Path does not exist: {}".format(path)
    if ext == "":
        if include_dirs:
            only_files = [os.path.join(os.path.abspath(path), f) for f in
                          os.listdir(path) if
                          os.path.exists(os.path.join(os.path.abspath(path), f))]
        else:
            only_files = [os.path.join(os.path.abspath(path), f) for f in
                          os.listdir(path) if
                          os.path.isfile(os.path.join(os.path.abspath(path), f))]
    else:
        only_files = [os.path.join(os.path.abspath(path), f) for f in
                      os.listdir(path) if
                      os.path.isfile(os.path.join(os.path.abspath(path), f))
                      if f.endswith(ext)]
    return only_files


def load_json(path):
    """Load a json file and make sure that path exists"""
    path = os.path.abspath(path)
    assert os.path.isfile(path), "Json file does not exist: {}".format(path)
    with open(path) as json_file:
        args = json.load(json_file)
    return args


def save_json(dict1, path):
    """Save a python object as a json file"""
    path = os.path.abspath(path)
    with open(path, 'w') as outfile:
        json.dump(dict1, outfile, indent=2)
    assert os.path.isfile(path)
    return path


def time_it(func, *args):
    """Very basic timing function
    :param func: callable function
    :param args: arguments to pass to function
    :return: object returned from function, time to complete
    """
    assert callable(func), "Function is not callable"
    start = datetime.datetime.now()
    something = func(*args)
    end = datetime.datetime.now()
    return something, (end - start).total_seconds()


def debug(verbose=False):
    """Method for setting log statements with verbose or not verbose"""
    assert type(verbose) is bool, "Verbose needs to be a boolean"
    if verbose:
        logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.DEBUG)
    else:
        logging.basicConfig(format="%(levelname)s: %(message)s")
        logging.info("This should not print.")


def check_numpy_table(data, req_fields):
    """CHeck if numpy table has required columns, raises errors if it does not

    :param data: numpy array
    :param req_fields: required fields to be found in numpy array
    :return True if passes test
    """
    if not isinstance(data, np.ndarray):
        raise TypeError('Data is not ndarray.')
    if data.dtype.names is not None:
        if not set(req_fields).issubset(data.dtype.names):
            raise KeyError(
                'Read data does not contain required fields: {}, got {}.'.format(
                    req_fields, data.dtype.names
                )
            )
    else:
        raise KeyError("Read data does not contain required fields")
    return True


def merge_two_dicts(dict1, dict2):
    """Given two dicts, merge them into a new dict as a shallow copy.
    source: https://stackoverflow.com/questions/38987/
    how-to-merge-two-python-dictionaries-in-a-single-expression"""
    assert type(dict1) is dict or type(dict1) is DotDict, "Both arguments must be dictionaries: type(arg1) = {}".format(
        type(dict1))
    assert type(dict2) is dict or type(dict2) is DotDict, "Both arguments must be dictionaries: type(arg2) = {}".format(
        type(dict2))
    final = dict1.copy()
    final.update(dict2)
    return final


def merge_dicts(dicts):
    """Given a list of dicts, merge them into a new dict.

    :param dicts: list of dictionaries to merge together
    """
    assert type(dicts) is list, "Must past list of dictionaries"
    for dictionary in dicts:
        assert type(dictionary) is dict or type(dictionary) is DotDict, "Every itme in dictionary must be a dictionary"

    final = dicts[0].copy()
    for dictionary in dicts[1:]:
        final.update(dictionary)
    return final


class TimeStamp:
    def __init__(self):
        self.timestamp = 0
        self.date = 0

    def posix_timestamp(self):
        now = datetime.datetime.now()
        now = now.timestamp()
        self.timestamp = now
        return self.timestamp

    @staticmethod
    def get_posix_date(timestamp):
        return str(datetime.datetime.utcfromtimestamp(timestamp))

    @staticmethod
    def get_local_date(timestamp):
        return str(datetime.datetime.fromtimestamp(timestamp))

    @staticmethod
    def str_to_timestamp(date, date_format='%Y-%m-%d %H:%M:%S.%f'):
        """Get timestamp from posix date"""
        date2 = datetime.datetime.strptime(date, date_format)
        timestamp = date2.replace(tzinfo=datetime.timezone.utc).timestamp()
        return timestamp

    def posix_date(self):
        if self.timestamp == 0:
            self.posix_timestamp()
        self.date = datetime.datetime.utcfromtimestamp(self.timestamp)
        return str(self.date)


def change_np_field_type(np_struct_array, name, data_type):
    """Return a new array that is like the structured numpy array, but has additional fields.
    description looks like description=[('test', '<i8')]

    :param np_struct_array: structured numpy array
    :param name: name of field in np_struct_array to change dtype
    :param data_type: dtype to change into
    """
    assert np_struct_array.dtype.fields is not None, "Must be a structured numpy array"
    assert name in np_struct_array.dtype.fields, "Name must be a field in the structured numpy array"
    dtypes = []
    # find field to replace
    for field in np_struct_array.dtype.descr:
        if field[0] == name:
            dtypes.append((name, data_type))
        else:
            dtypes.append(field)
    # create new array and copy fields over
    new = np.zeros(np_struct_array.shape, dtype=dtypes)
    for name in np_struct_array.dtype.names:
        new[name] = np_struct_array[name]
    return new


def allLexicographicRecur(string, data, last, index):
    """Generator for all permutations of an alphabetically sorted string

    source: https://www.geeksforgeeks.org/print-all-permutations-with-repetition-of-characters/
    :param string: alphabetically sorted string
    :param data: list of characters
    :param last: length of string wanted to iterate
    :param index: current index to fix
    :return: generator of strings
    """
    length = len(string)

    # One by one fix all characters at the given index and
    # recur for the subsequent indexes
    for i in range(length):

        # Fix the ith character at index and if this is not
        # the last index then recursively call for higher
        # indexes
        data[index] = string[i]

        # If this is the last index then print the string
        # stored in data[]
        if index == last:
            yield ''.join(data)
        else:
            for x in allLexicographicRecur(string, data, last, index + 1):
                yield x


def all_string_permutations(string, length=None):
    """Generates all lexicographically sorted permutations of a string

    source: https://www.geeksforgeeks.org/print-all-permutations-with-repetition-of-characters/
    :param string: string to create all lexicographically sorted permutations of a string
    :param length: length of desired strings
    :return: generator of strings
    """
    if length is None:
        length = len(string)
    assert length > 0, "Cannot provide empty string or length under 1"

    # Create a temp array that will be used by
    # allLexicographicRecur()
    data = [""] * (length + 1)

    # Sort the input string so that we get all output strings in
    # lexicographically sorted order
    string = sorted(''.join(set(string)))

    # Now print all permutaions
    return allLexicographicRecur(string, data, length - 1, 0)


def get_random_string(length, chars=string.ascii_uppercase):
    """Get random strings with given characters and given length

    note: if duplicate characters given, they will not be removed. It is assumed that the user wants certain characters
            selected more frequently than other characters

    :param length: length of random string
    :param chars: chars to select randomly, default is all uppercase ascii chars.
    """
    return ''.join(random.choice(chars) for _ in range(length))


def get_random_strings(n_strings, len_string, chars=string.ascii_uppercase):
    """Generate N unique random strings given an alphabet and constant length of the string
    :param n_strings: number of strings to generate
    :param len_string: size of the strings
    :param chars: alphabet to work with
    """
    assert max_number_of_possible_strings(len_string, len(chars)) >= n_strings, \
        "Cannot generate this many random strings." \
        " max_n_possible_strings {} > n_strings {}".format(max_number_of_possible_strings(len_string, len(chars)),
                                                           n_strings)

    if max_number_of_possible_strings(len_string, len(chars)) / 2 <= n_strings:
        all_strings = np.random.choice([x for x in all_string_permutations(chars, len_string)], n_strings,
                                       replace=False)
    else:
        all_strings = set()
        while len(all_strings) < n_strings:
            all_strings |= {get_random_string(len_string, chars)}

    return list(all_strings)


def max_number_of_possible_strings(len_string, n_chars):
    return n_chars ** len_string


def find_substring_indices(main_string, sub_string, offset=0, overlap=True):
    """Generator that finds start positions of specific substring in a string, including overlapping substrings

    :param main_string: main_string to search
    :param sub_string: string to find instances of within the main_string
    :param offset: if you are searching for the index of an internal character of the substring we will add the
                    offset to the index. Default: 0
    :param overlap: boolean option if you want to find overlap eg.                Default: True
                    find_substring_indices("AAAA", "AA", overlap=True) = [0,1,2]
                    find_substring_indices("AAAA", "AA", overlap=False) = [0,2]

    :return: generator yielding (substring_start_index + offset)
    """
    assert len(sub_string) >= 1, "You must select a substring with len(sub_string) >= 1: {}".format(sub_string)
    assert len(main_string) >= 1, "You must select a main_string with len(main_string) >= 1: {}".format(main_string)
    if overlap:
        sub = '(?=(' + sub_string + "))"
    else:
        sub = sub_string
    for m in re.finditer(sub, main_string):
        yield m.start() + offset


def merge_lists(list_of_lists):
    """merge all items from a list of lists into a single list
    :param list_of_lists: a list of lists
    :return: a list with all elements from all lists
    """
    return [item for sublist in list_of_lists for item in sublist]


def count_lines_in_file(txt_file):
    """Count the lines in a text file

    source: https://stackoverflow.com/questions/845058/how-to-get-line-count-cheaply-in-python

    :param txt_file: path to text file of any type
    :return: number of lines in file
    """
    assert os.path.isfile(txt_file), "Input file is not a file. {}".format(txt_file)
    f = open(txt_file, 'rb')
    bufgen = takewhile(lambda x: x, (f.raw.read(1024 * 1024) for _ in repeat(None)))
    total = sum(buf.count(b'\n') for buf in bufgen)
    f.close()
    return total


def all_abspath(test_path):
    """Return original object if it is not string

    :param test_path: preferably a string object but will not error if non string object is passed
    """
    if type(test_path) is str:
        return os.path.abspath(test_path)
    else:
        return test_path


def split_every(n, iterable):
    """https://stackoverflow.com/questions/9475241/split-string-every-nth-character"""
    i = iter(iterable)
    piece = list(islice(i, n))
    while piece:
        yield piece
        piece = list(islice(i, n))


def split_every_string(n, string):
    """https://stackoverflow.com/questions/9475241/split-string-every-nth-character"""
    for x in split_every(n, string):
        yield "".join(x)


def list_dir_recursive(directory, ext=""):
    """Dive into directory and yield all files with given extension"""
    for root, dirs, files in os.walk(directory):
        for file in files:
            if file.endswith(ext):
                yield os.path.join(root, file)


def get_all_sub_directories(directory, hidden=False):
    """Get all sub directories from a parent directory"""
    for root, dirs, files in os.walk(directory):
        for dir1 in dirs:
            if not hidden:
                if not (dir1.startswith(".") or dir1.startswith("_")):
                    yield os.path.join(root, dir1)
            else:
                yield os.path.join(root, dir1)


def binary_search(sorted_list, target, exact_match=True):
    """Binary search through a sorted list of numbers for a target. If exact_match is set to true, it will only return
        if target is in sorted_list
    :param sorted_list:
    :param target:
    :param exact_match:
    :return: index of item or

    """
    min = 0
    max = len(sorted_list) - 1
    avg = int(np.round((min + max) / 2))
    if exact_match:
        def return_condition():
            return sorted_list[avg] == target
    else:
        def return_condition():
            if avg == min:
                if target < sorted_list[avg]:
                    return True
            if avg == max:
                if sorted_list[avg] < target:
                    return True
            else:
                return sorted_list[avg] < target < sorted_list[avg + 1]

    while min < max:
        if return_condition():
            return avg
        elif sorted_list[avg] < target:
            return avg + 1 + binary_search(sorted_list[avg + 1:], target, exact_match)
        else:
            return binary_search(sorted_list[:avg], target, exact_match)

    # avg may be a partial offset so no need to print it here
    # print "The location of the number in the array is", avg
    return avg


def tar_gz(source_dir, out_path, keep_structure=True):
    """Create a tar file from a directory of files
    :param source_dir: path to directory to tar gz
    :param out_path: path to output file or directory
    :param keep_structure: keep directory structure
    """
    assert os.path.isdir(source_dir), "Source directory must be a directory: {}".format(source_dir)
    if keep_structure:
        arcname = os.path.basename(source_dir)
    else:
        arcname = None

    if os.path.isdir(out_path):
        out_path = os.path.join(out_path, os.path.basename(source_dir)+".tar.gz")

    with tarfile.open(out_path, "w:gz") as tar:
        tar.add(source_dir, arcname=arcname)

    return out_path


def untar_gz(tar_gz_file, out_dir="."):
    """Create a tar file from a directory of files
    :param tar_gz_file: tar.gz file to untar and unzip
    :param out_dir: path to directory to unpack tar file
    """
    tar = tarfile.open(tar_gz_file, "r:gz")
    tar.extractall(path=out_dir)
    tar.close()
    return out_dir


if __name__ == '__main__':
    print("This is a library of python functions")



#!/usr/bin/env python
"""Helper functions to multiprocess functions easily"""
########################################################################
# File: multiprocess.py
#  executable: multiprocess.py
#
# Author: Andrew Bailey, Most work was originally done by Trevor Pesout
# History: 03/05/18 Created
########################################################################

import time
from multiprocessing import Manager, Process, current_process
from timeit import default_timer as timer

TOTAL_KEY = "total"
FAILURE_KEY = "failure"
MEM_USAGE_KEY = "mem_usage"


def run_service(service, iterable, arguments, iterable_argument_names, worker_count,
                service_arguments={}, log_function=print):
    """Take in a service (function) with an iterable list, arguments to be passed to
        the "work" function done on each element, and an argument to append to these 'work' arguments as name of
        the element.

        :param service: function which will extract arguments from a work queue and put return values in done queue
        :param iterable: iterable list of arguments to be passed in
        :param arguments: all non iterable arguments
        :param iterable_argument_names: names of the arguments which are being zipped with the
        :param worker_count: number of workers or processes to spin up
        :param service_arguments: arguments to be passed to Process
        :param log_function: usually just print but will pass logging string info if log function is used
        :return: total, failure, messages, output
    """

    start = timer()
    args = list(arguments.keys())
    args.extend(iterable_argument_names)
    if log_function is not None:
        log_function("[run_service] running service {} with {} workers".format(service, worker_count))

    # setup workers for multiprocessing
    work_queue = Manager().Queue()
    done_queue = Manager().Queue()

    # add everything to work queue
    jobs = []
    for x in iterable:
        if type(x) is not tuple:
            x = [x]
        args = dict(dict(zip(iterable_argument_names, x)),
                    **arguments)
        work_queue.put(args)

    # start workers
    for w in range(worker_count):
        p = Process(target=service, args=(work_queue, done_queue), kwargs=service_arguments)
        p.start()
        jobs.append(p)
        work_queue.put('STOP')

    # wait for threads to finish, then stop the done queue
    for p in jobs:
        p.join()
    done_queue.put('STOP')

    # if example service model is used, metrics can be gathered in this way
    messages = []
    output = []
    total = 0
    failure = 0
    no_stop = True
    while no_stop:
        f = done_queue.get()
        if isinstance(f, str):
            if f == "STOP":
                no_stop = False
            elif f.startswith(TOTAL_KEY):
                total += int(f.split(":")[1])
            elif f.startswith(FAILURE_KEY):
                failure += int(f.split(":")[1])
            else:
                messages.append(f)
        else:
            output.append(f)
    stop = timer()
    log_function("[run_service] Completed in {} seconds.".format(stop - start))
    # if we should be logging and if there is material to be logged
    if log_function is not None and (total + failure + len(messages)) > 0:
        log_function(
            "[run_service] Summary {}:\n[run_service]\tTime: {}s\n[run_service]\tTotal: {}\n[run_service]\tFailure: {}"
            .format(service, time.time() - start, total, failure))
        log_function("[run_service]\tMessages:\n[run_service]\t\t{}".format("\n[run_service]\t\t".join(messages)))
    # return relevant info
    return total, failure, messages, output


class BasicService(object):

    def __init__(self, function):
        self.function = function

    def run(self, work_queue, done_queue, service_name="example_service"):

        """
        Service used by the multiprocess module for an example of what should be created for multiprocessing
        :param work_queue: arguments to be done
        :param done_queue: errors and returns to be put
        :param service_name: name of the service
        """
        # prep
        total_handled = 0
        failure_count = 0

        # catch overall exceptions
        try:
            for f in iter(work_queue.get, 'STOP'):
                # catch exceptions on each element
                try:
                    return_value = self.function(**f)
                    done_queue.put(return_value)
                except Exception as e:
                    # get error and log it
                    message = "{}:{}".format(type(e), str(e))
                    error = "{} '{}' failed with: {}".format(service_name, current_process().name, message)
                    print("[{}] ".format(service_name) + error)
                    done_queue.put(error)
                    failure_count += 1

                # increment total handling
                total_handled += 1

        except Exception as e:
            # get error and log it
            message = "{}:{}".format(type(e), str(e))
            error = "{} '{}' critically failed with: {}".format(service_name, current_process().name, message)
            print("[{}] ".format(service_name) + error)
            done_queue.put(error)

        finally:
            # logging and final reporting
            print("[%s] '%s' completed %d calls with %d failures"
                  % (service_name, current_process().name, total_handled, failure_count))
            done_queue.put("{}:{}".format(TOTAL_KEY, total_handled))
            done_queue.put("{}:{}".format(FAILURE_KEY, failure_count))



#!/usr/bin/env python
"""Classification output analysis functions and classes"""
########################################################################
# File: classification.py
#  executable: classification.py
#
# Author: Andrew Bailey
# History: 12/19/18 Created
########################################################################

import os
import pandas as pd
import numpy as np
import platform
import matplotlib as mpl

if os.environ.get('DISPLAY', '') == '':
    print('no display found. Using non-interactive Agg backend')
    mpl.use('Agg')
if platform.system() == "Darwin":
    mpl.use("macosx")
import matplotlib.pyplot as plt
from itertools import cycle
from sklearn.metrics import roc_curve, auc, confusion_matrix, average_precision_score, brier_score_loss, \
    precision_recall_curve
from sklearn.calibration import calibration_curve
from sklearn.utils import column_or_1d, assert_all_finite, check_consistent_length
from sklearn.utils.extmath import stable_cumsum
from sklearn.utils.fixes import signature
from scipy import interp
from py3helpers.utils import binary_search
from itertools import cycle


class ClassificationMetrics(object):

    def __init__(self, binary_labels, class_probabilities, class_ns=None, label_ids=None):
        """Initialize binary labels and class probabilities"""
        assert isinstance(binary_labels, pd.DataFrame), \
            "binary_labels must be a pandas dataframe: type {}".format(type(binary_labels))
        assert isinstance(class_probabilities, pd.DataFrame), \
            "class_probabilities must be a pandas dataframe: type {}".format(type(class_probabilities))

        self.fps = dict()
        self.tps = dict()
        self.fns = dict()
        self.tns = dict()

        self.fpr = dict()
        self.tpr = dict()
        self.fnr = dict()
        self.tnr = dict()
        self.ppv = dict()
        self.fdr = dict()
        self.for1 = dict()
        self.npv = dict()
        self.thresholds = dict()
        self.roc_auc = dict()
        self.average_precision = dict()
        self.sorted_ids = dict()

        self.total = 0
        self.class_totals = dict()
        if label_ids is not None:
            assert len(label_ids) == len(binary_labels), \
                "label_ids length must match the length of binary labels. " \
                "len(label_ids) = {}, len(binary_labels) = {}".format(len(label_ids), len(binary_labels))

        self.label_ids = label_ids
        self.binary_labels = binary_labels
        self.class_probabilities = class_probabilities
        assert len(self.binary_labels) > 1, "binary_labels need to be longer than 1: {}".format(self.binary_labels)
        assert len(self.class_probabilities) > 1, "class_probabilities need to be longer than 1: {}".format(
            self.class_probabilities)

        self.class_ns = class_ns
        self.n_classes = 0
        self._set_check_class_ns()
        self._calculate_fp_tp_rate()

    def _set_check_class_ns(self):
        """Check class names from input data and optional argument

        Note: if class_ns is not passed in then we will will throw error
        """
        if self.class_ns is None:
            tmp_class_ns = [str(x) for x in self.binary_labels.dtypes.index]
            tmp_class_ns_2 = [str(x) for x in self.class_probabilities.dtypes.index]
            assert tmp_class_ns == tmp_class_ns_2, \
                "binary_labels and class_probabilities must have same column names: " \
                "binary_labels - {}, class_probabilities - {}".format(tmp_class_ns, tmp_class_ns_2)
            self.class_ns = tmp_class_ns
        assert len(self.class_ns) == self.binary_labels.shape[1]
        assert len(self.class_ns) == self.class_probabilities.shape[1]
        self.n_classes = len(self.class_ns)

    def _calculate_fp_tp_rate(self):
        """Calculate false positives and true positive rates"""
        # Compute false and true positives, ROC data and other important binary classification metrics
        for class_n in self.class_ns:
            self.fps[class_n], self.tps[class_n], self.thresholds[class_n], self.sorted_ids[class_n] = \
                binary_clf_curve_with_ids(self.binary_labels[class_n], self.class_probabilities[class_n],
                                          y_true_ids=self.label_ids)
            if self.tps[class_n].size == 0 or self.fps[class_n][0] != 0 or self.tps[class_n][0] != 0:
                # Add an extra threshold position if necessary
                self.tps[class_n] = np.r_[0, self.tps[class_n]]
                self.fps[class_n] = np.r_[0, self.fps[class_n]]
                self.sorted_ids[class_n] = np.r_[None, self.sorted_ids[class_n]]
                self.thresholds[class_n] = np.r_[self.thresholds[class_n][0] + 1, self.thresholds[class_n]]

            self.fns[class_n] = self.tps[class_n][-1] - self.tps[class_n]
            self.tns[class_n] = self.fps[class_n][-1] - self.fps[class_n]
            # positive predictive value
            self.ppv[class_n] = self.tps[class_n] / (self.tps[class_n] + self.fps[class_n])
            # false discovery rate
            self.fdr[class_n] = self.fps[class_n] / (self.tps[class_n] + self.fps[class_n])
            # negative predictive value
            self.npv[class_n] = self.tns[class_n] / (self.tns[class_n] + self.fns[class_n])
            # false omission rate
            self.for1[class_n] = self.fns[class_n] / (self.tns[class_n] + self.fns[class_n])

            # true/false positive/negative rates
            self.fpr[class_n] = self.fps[class_n] / self.fps[class_n][-1]
            self.tpr[class_n] = self.tps[class_n] / self.tps[class_n][-1]
            self.tnr[class_n] = 1 - self.fpr[class_n]
            self.fnr[class_n] = 1 - self.tpr[class_n]
            self.roc_auc[class_n] = auc(self.fpr[class_n], self.tpr[class_n])
            self.average_precision[class_n] = average_precision_score(self.binary_labels[class_n],
                                                                      self.class_probabilities[class_n])

        self.fpr["roc_micro"], self.tpr["roc_micro"], self.thresholds["roc_micro"] = \
            roc_curve(self.binary_labels.values.ravel(), self.class_probabilities.values.ravel())
        self.tnr["roc_micro"] = 1 - self.fpr["roc_micro"]
        self.fnr["roc_micro"] = 1 - self.tpr["roc_micro"]
        self.roc_auc["roc_micro"] = auc(self.fpr["roc_micro"], self.tpr["roc_micro"])
        self.average_precision["pr_micro"] = average_precision_score(self.binary_labels.values.ravel(),
                                                                     self.class_probabilities.values.ravel())
        self.ppv["pr_micro"], self.tpr["pr_micro"], _ = precision_recall_curve(self.binary_labels.values.ravel(),
                                                                               self.class_probabilities.values.ravel())

        all_fpr = np.unique(np.concatenate([self.fpr[i] for i in self.class_ns]))

        # Then interpolate all ROC curves at this points
        mean_tpr = np.zeros_like(all_fpr)
        for class_n in self.class_ns:
            mean_tpr += interp(all_fpr, self.fpr[class_n], self.tpr[class_n])

        # Finally average it and compute AUC
        mean_tpr /= self.n_classes

        self.fpr["macro"] = all_fpr
        self.tpr["macro"] = mean_tpr
        self.tnr["macro"] = 1 - self.fpr["macro"]
        self.fnr["macro"] = 1 - self.tpr["macro"]

        self.roc_auc["macro"] = auc(self.fpr["macro"], self.tpr["macro"])

    def _get_index_from_threshold(self, class_n, threshold):
        """Get correct index from any arbitrary probability threshold cutoff"""
        assert class_n in self.class_ns, \
            "Class name is not in class names. {} not in {}".format(class_n, self.class_ns)
        index = binary_search(self.thresholds[class_n][::-1], threshold, exact_match=False)
        if index == len(self.thresholds[class_n]) - 1:
            return 0
        elif index == 0:
            return len(self.thresholds[class_n]) - 1
        else:
            return len(self.thresholds[class_n]) - index - 2

    def recall(self, class_n, threshold):
        return self.true_positive_rate(class_n, threshold)

    def sensitivity(self, class_n, threshold):
        return self.true_positive_rate(class_n, threshold)

    def specificity(self, class_n, threshold):
        return self.true_negative_rate(class_n, threshold)

    def true_positive_rate(self, class_n, threshold):
        return self.tpr[class_n][self._get_index_from_threshold(class_n, threshold)]

    def false_positive_rate(self, class_n, threshold):
        return self.fpr[class_n][self._get_index_from_threshold(class_n, threshold)]

    def true_negative_rate(self, class_n, threshold):
        return self.tnr[class_n][self._get_index_from_threshold(class_n, threshold)]

    def false_negative_rate(self, class_n, threshold):
        return self.fnr[class_n][self._get_index_from_threshold(class_n, threshold)]

    def positive_likelihood_ratio(self, class_n, threshold):
        fpr = self.false_positive_rate(class_n, threshold)
        tpr = self.true_positive_rate(class_n, threshold)
        return tpr / fpr

    def negative_likelihood_ratio(self, class_n, threshold):
        fpr = self.false_positive_rate(class_n, threshold)
        tpr = self.true_positive_rate(class_n, threshold)
        return fpr / tpr

    def diagnostic_odds_ratio(self, class_n, threshold):
        plr = self.positive_likelihood_ratio(class_n, threshold)
        nlr = self.negative_likelihood_ratio(class_n, threshold)
        return plr / nlr

    def positive_predictive_value(self, class_n, threshold):
        return self.ppv[class_n][self._get_index_from_threshold(class_n, threshold)]

    def precision(self, class_n, threshold):
        return self.positive_predictive_value(class_n, threshold)

    def accuracy(self, class_n, threshold):
        return (self.get_n_tps(class_n, threshold) + self.get_n_tns(class_n, threshold)) / len(self.binary_labels)

    def false_discovery_rate(self, class_n, threshold):
        return self.fdr[class_n][self._get_index_from_threshold(class_n, threshold)]

    def negative_predictive_value(self, class_n, threshold):
        return self.npv[class_n][self._get_index_from_threshold(class_n, threshold)]

    def false_omission_rate(self, class_n, threshold):
        return self.for1[class_n][self._get_index_from_threshold(class_n, threshold)]

    def f1_score(self, class_n, threshold):
        precision = self.precision(class_n, threshold)
        recall = self.recall(class_n, threshold)
        return 2 / ((1 / precision) + (1 / recall))

    def prevalence(self, class_n):
        return np.count_nonzero(self.binary_labels[class_n]) / len(self.binary_labels)

    def confusion_matrix(self):
        labels = self.binary_labels.idxmax(1)
        predictions = self.class_probabilities.idxmax(1)
        return confusion_matrix(labels, predictions, labels=self.class_ns)

    def plot_calibration_curve(self, class_n, save_fig_path=None):
        if save_fig_path is not None:
            assert os.path.exists(os.path.dirname(save_fig_path)), \
                "Output directory does not exist: {}".format(save_fig_path)

        labels = self.binary_labels[class_n]
        predictions = self.class_probabilities[class_n]
        clf_score = brier_score_loss(labels, predictions)
        fraction_of_positives, mean_predicted_value = \
            calibration_curve(labels, predictions, n_bins=10)

        plt.figure(figsize=(10, 10))
        ax1 = plt.subplot2grid((3, 1), (0, 0), rowspan=2)
        ax2 = plt.subplot2grid((3, 1), (2, 0))

        ax1.plot([0, 1], [0, 1], "k:", label="Perfectly calibrated")

        ax1.plot(mean_predicted_value, fraction_of_positives, "s-",
                 label="%s (%1.3f)" % (class_n, clf_score))

        ax2.hist(predictions, range=(0, 1), bins=10, label=class_n,
                 histtype="step", lw=2)
        ax1.set_ylabel("Fraction of positives")
        ax1.set_ylim([-0.05, 1.05])
        ax1.legend(loc="lower right")
        ax1.set_title('Calibration plots  (reliability curve)')

        ax2.set_xlabel("Mean predicted value")
        ax2.set_ylabel("Count")
        ax2.legend(loc="upper center", ncol=2)
        if save_fig_path is not None:
            plt.savefig(save_fig_path)
        else:
            plt.show()
            pass

    def get_fp_ids(self, class_n, threshold=0.5):
        """Get the false positive ids given a class and threshold"""
        return self._get_ids(self.fps, class_n, threshold)

    def get_tp_ids(self, class_n, threshold=0.5):
        """Get the false positive ids given a class and threshold"""
        return self._get_ids(self.tps, class_n, threshold)

    def get_fn_ids(self, class_n, threshold=0.5):
        """Get the false positive ids given a class and threshold"""
        return self._get_ids2(self.fns, class_n, threshold)

    def get_tn_ids(self, class_n, threshold=0.5):
        """Get the false positive ids given a class and threshold"""
        return self._get_ids2(self.tns, class_n, threshold)

    def get_n_tps(self, class_n, threshold=0.5):
        """Get number of false positives"""
        return self._get_n(self.tps, class_n, threshold=threshold)

    def get_n_fps(self, class_n, threshold=0.5):
        """Get number of false positives"""
        return self._get_n(self.fps, class_n, threshold=threshold)

    def get_n_tns(self, class_n, threshold=0.5):
        """Get number of false positives"""
        return self._get_n(self.tns, class_n, threshold=threshold)

    def get_n_fns(self, class_n, threshold=0.5):
        """Get number of false positives"""
        return self._get_n(self.fns, class_n, threshold=threshold)

    def _get_n(self, place, class_n, threshold=0.5):
        """Get number of false positives"""
        return place[class_n][self._get_index_from_threshold(class_n, threshold=threshold)]

    def _get_ids(self, place, class_n, threshold=0.5):
        """Get the false and true positive ids given a class and threshold"""
        assert self.label_ids is not None, "Cannot get ids if you do not pass in label_ids to class."
        index = self._get_index_from_threshold(class_n, threshold)
        data = place[class_n]
        names = []
        prev_n_calls = 0
        for x in range(index + 1):
            n_calls = data[x]
            if 0 < n_calls != prev_n_calls:
                if x >= len(self.sorted_ids[class_n]):
                    print("Wait what?")
                names.append(self.sorted_ids[class_n][x])
            prev_n_calls = n_calls
        return names

    def _get_ids2(self, place, class_n, threshold=0.5):
        """Get the false and true negative ids given a class and threshold"""
        assert self.label_ids is not None, "Cannot get ids if you do not pass in label_ids to class."
        index = self._get_index_from_threshold(class_n, threshold)
        data = place[class_n]
        names = []
        prev_n_calls = 0
        for x in reversed(range(index, len(place[class_n]))):
            n_calls = data[x]
            if n_calls != prev_n_calls:
                if x >= len(self.sorted_ids[class_n]):
                    print("Wait what?")
                names.append(self.sorted_ids[class_n][x + 1])
            prev_n_calls = n_calls
        return names

    def plot_roc(self, class_n, save_fig_path=None, title="Receiver operating characteristic"):
        if save_fig_path is not None:
            assert os.path.exists(os.path.dirname(save_fig_path)), \
                "Output directory does not exist: {}".format(save_fig_path)

        plt.figure()
        lw = 2
        plt.plot(self.fpr[class_n], self.tpr[class_n], color='darkorange',
                 lw=lw, label='ROC curve (area = %0.2f)' % self.roc_auc[class_n])
        plt.plot([0, 1], [0, 1], color='navy', lw=lw, linestyle='--')
        plt.xlim([0.0, 1.0])
        plt.ylim([0.0, 1.05])
        plt.xlabel('False Positive Rate')
        plt.ylabel('True Positive Rate')
        plt.title(title)
        plt.legend(loc="lower right")
        if save_fig_path is not None:
            plt.savefig(save_fig_path)
        else:
            plt.show()
            pass

    def plot_multiclass_roc(self, lw=2, save_fig_path=None):
        # Plot all ROC curves
        if save_fig_path is not None:
            assert os.path.exists(os.path.dirname(save_fig_path)), \
                "Output directory does not exist: {}".format(save_fig_path)

        plt.figure()
        plt.plot(self.fpr["roc_micro"], self.tpr["roc_micro"],
                 label='micro-average ROC curve (area = {0:0.2f})'
                       ''.format(self.roc_auc["roc_micro"]),
                 color='deeppink', linestyle=':', linewidth=4)

        plt.plot(self.fpr["macro"], self.tpr["macro"],
                 label='macro-average ROC curve (area = {0:0.2f})'
                       ''.format(self.roc_auc["macro"]),
                 color='navy', linestyle=':', linewidth=4)

        colors = cycle(['aqua', 'darkorange', 'cornflowerblue'])
        for i, color in zip(self.class_ns, colors):
            plt.plot(self.fpr[i], self.tpr[i], color=color, lw=lw,
                     label='ROC curve of class {0} (area = {1:0.2f})'
                           ''.format(i, self.roc_auc[i]))

        plt.plot([0, 1], [0, 1], 'k--', lw=lw)
        plt.xlim([0.0, 1.0])
        plt.ylim([0.0, 1.05])
        plt.xlabel('False Positive Rate')
        plt.ylabel('True Positive Rate')
        plt.title('Some extension of Receiver operating characteristic to multi-class')
        plt.legend(loc="lower right")
        if save_fig_path is not None:
            plt.savefig(save_fig_path)
        else:
            plt.show()
            pass

    def plot_precision_recall(self, class_n, save_fig_path=None, title="Precision Recall curve"):
        if save_fig_path is not None:
            assert os.path.exists(os.path.dirname(save_fig_path)), \
                "Output directory does not exist: {}".format(save_fig_path)

        plt.figure()
        lw = 2
        plt.plot(self.tpr[class_n], self.ppv[class_n], color='darkblue',
                 lw=lw, label='Precision Recall curve (Average Precision = %0.2f)' % self.average_precision[class_n])
        plt.fill_between(self.tpr[class_n], self.ppv[class_n], alpha=0.2, color='darkblue')

        plt.xlim([0.0, 1.0])
        plt.ylim([0.0, 1.05])
        plt.xlabel('Recall (Positive predictive Value')
        plt.ylabel('Precision (True Positive Rate')
        plt.title(title)
        plt.legend(loc="lower right")
        if save_fig_path is not None:
            plt.savefig(save_fig_path)
        else:
            plt.show()
            pass

    def plot_micro_average_precision_score(self, save_fig_path=None):
        if save_fig_path is not None:
            assert os.path.exists(os.path.dirname(save_fig_path)), \
                "Output directory does not exist: {}".format(save_fig_path)

        step_kwargs = ({'step': 'post'}
                       if 'step' in signature(plt.fill_between).parameters
                       else {})

        plt.figure()
        plt.step(self.tpr['pr_micro'], self.ppv['pr_micro'], color='b', alpha=0.2,
                 where='post')
        plt.fill_between(self.tpr["pr_micro"], self.ppv["pr_micro"], alpha=0.2, color='b',
                         **step_kwargs)

        plt.xlabel('Recall')
        plt.ylabel('Precision')
        plt.ylim([0.0, 1.05])
        plt.xlim([0.0, 1.0])
        plt.title('Average precision score, micro-averaged over all classes: '
                  'AP={0:0.2f}'.format(self.average_precision["pr_micro"]))
        if save_fig_path is not None:
            plt.savefig(save_fig_path)
        else:
            plt.show()
            pass

    def plot_multiclass_precision_recall(self, save_fig_path=None):
        if save_fig_path is not None:
            assert os.path.exists(os.path.dirname(save_fig_path)), \
                "Output directory does not exist: {}".format(save_fig_path)

        # setup plot details
        colors = cycle(['navy', 'turquoise', 'darkorange', 'cornflowerblue', 'teal'])

        plt.figure(figsize=(7, 8))
        f_scores = np.linspace(0.2, 0.8, num=4)
        lines = []
        labels = []
        for f_score in f_scores:
            x = np.linspace(0.01, 1)
            y = f_score * x / (2 * x - f_score)
            l, = plt.plot(x[y >= 0], y[y >= 0], color='gray', alpha=0.2)
            plt.annotate('f1={0:0.1f}'.format(f_score), xy=(0.9, y[45] + 0.02))

        lines.append(l)
        labels.append('iso-f1 curves')
        l, = plt.plot(self.tpr["pr_micro"], self.ppv["pr_micro"], color='gold', lw=2)
        lines.append(l)
        labels.append('micro-average Precision-recall (area = {0:0.2f})'
                      ''.format(self.average_precision["pr_micro"]))

        for class_name, color in zip(self.class_ns, colors):
            l, = plt.plot(self.tpr[class_name], self.ppv[class_name], color=color, lw=2)
            lines.append(l)
            labels.append('Precision-recall for class {0} (area = {1:0.2f})'
                          ''.format(class_name, self.average_precision[class_name]))

        fig = plt.gcf()
        fig.subplots_adjust(bottom=0.25)
        plt.xlim([0.0, 1.0])
        plt.ylim([0.0, 1.05])
        plt.xlabel('Recall')
        plt.ylabel('Precision')
        plt.title('Extension of Precision-Recall curve to multi-class')
        plt.legend(lines, labels, loc=(0, -.38), prop=dict(size=14))

        if save_fig_path is not None:
            plt.savefig(save_fig_path)
        else:
            plt.show()
            pass

    def get_confusion_matrix(self, threshold=0.5):
        """Get confusion matrix from binary classification
        :param threshold: threshold for cutoff between classes
        """
        assert len(self.class_ns) == 2, "Only two classes can be used for get_confusion_matrix"
        class_n = self.class_ns[0]
        tp = self.tps[class_n][self._get_index_from_threshold(class_n, threshold)]
        fn = self.fns[class_n][self._get_index_from_threshold(class_n, threshold)]
        tn = self.tns[class_n][self._get_index_from_threshold(class_n, threshold)]
        fp = self.fps[class_n][self._get_index_from_threshold(class_n, threshold)]

        false_discovery_rate = fp / (tp + fp)
        precision = tp / (tp + fp)
        false_omission_rate = fn / (tn + fn)
        negative_predictive_value = tn / (tn + fn)

        true_positive_rate_recall = tp / (tp + fn)
        false_negative_rate = fn / (tp + fn)

        false_positive_rate = fp / (fp + tn)
        true_negative_rate_specificity = tn / (tn + fp)

        positive_likelihood_ratio = true_positive_rate_recall / false_positive_rate
        negative_likelihood_ratio = false_negative_rate / true_negative_rate_specificity

        diagnostic_odds_ratio = positive_likelihood_ratio / negative_likelihood_ratio

        f1_score = 2 * ((precision * true_positive_rate_recall) / (precision + true_positive_rate_recall))

        return (np.asarray(
            [[tp, fp, precision, false_discovery_rate],
             [fn, tn, false_omission_rate, negative_predictive_value],
             [true_positive_rate_recall, false_positive_rate, positive_likelihood_ratio, diagnostic_odds_ratio],
             [false_negative_rate, true_negative_rate_specificity, negative_likelihood_ratio, f1_score]]))

    def plot_confusion_matrix(self, threshold=0.5, title="Confusion Matrix", save_fig_path=None):
        """Plot the confusion matrix with the information of each box explained
        :param threshold: classification threshold to plot
        :param title: Title of the confusion matrix
        :param save_fig_path: path to save figure
        """
        assert len(self.class_ns) == 2, "Only two classes can be used to plot confusion matrix"
        class_n = self.class_ns[0]
        tp = self.get_n_tps(class_n, threshold)
        fn = self.get_n_fns(class_n, threshold)
        tn = self.get_n_tns(class_n, threshold)
        fp = self.get_n_fps(class_n, threshold)

        data = \
            np.asarray([[tp, fp,
                         "Precision\nPositive Predictive Value\ntp / (tp + fp)\n{0:.3f}".format(
                             self.positive_predictive_value(class_n, threshold)),
                         "False Discovery Rate\nfp / (tp + fp)\n{0:.3f}".format(self.false_discovery_rate(class_n,
                                                                                                          threshold))],
                        [fn, tn,
                         "False Omission Rate\nfn / (tn + fn)\n{0:.3f}".format(self.false_omission_rate(class_n,
                                                                                                        threshold)),
                         "Negative Predictive Value\ntn / (tn + fn)\n{0:.3f}".format(
                             self.negative_predictive_value(class_n, threshold))],
                        ["Recall\nTrue Positive Rate\ntp / (tp + fn)\n{0:.3f}".format(self.true_positive_rate(class_n,
                                                                                                              threshold)),
                         "False Positive Rate\nfp / (fp + tn)\n{0:.3f}".format(self.false_positive_rate(class_n,
                                                                                                        threshold)),
                         "Positive Likelihood Ratio\ntpr / fpr\n{0:.3f}".format(
                             self.positive_likelihood_ratio(class_n, threshold)),
                         "Diagnostic Odds Ratio\nplr / nlr \n{0:.3f}".format(self.diagnostic_odds_ratio(class_n, threshold))],
                        ["False Negative Rate\nfn / (tp + fn)\n{0:.3f}".format(self.false_negative_rate(class_n,
                                                                                                        threshold)),
                         "Specificity\nTrue Negative Rate\ntn / (tn + fp)\n{0:.3f}".format(
                             self.true_negative_rate(class_n, threshold)),
                         "Negative Likelihood Ratio\nfnr/tnr\n{0:.3f}".format(
                             self.negative_likelihood_ratio(class_n, threshold)),
                         "F1 Score\n2*[(ppv * tpr) / (ppv + tpr)]\n{0:.3f}".format(self.f1_score(class_n, threshold))]])

        fig = plt.figure(figsize=(15, 5))
        col_labels = ['Condition Positive (cp): {}'.format(self.class_ns[0]), "Condition Negative (cn): {}".format(self.class_ns[1]),
                      "Prevalence\n cp / total\n{0:.3f}".format(self.prevalence(class_n)),
                      "Accuracy\n[sum(tp) + sum(tn)] / total\n{0:.3f}".format(self.accuracy(class_n, threshold))]
        row_labels = ["Predicted Condition Positive: {}".format(self.class_ns[0]), "Predicted Condition Negative: {}".format(self.class_ns[1]), "", ""]
        table_vals = data
        cmap = mpl.cm.get_cmap('Blues')
        blank = (0, 0, 0, 0)
        total_p = sum([tp, fp])
        total_n = sum([tn, fn])
        colors = [[cmap(tp/total_p), cmap(fp/total_p), blank, blank], [cmap(fn/total_n), cmap(tn/total_n), blank, blank],
                  [blank]*4, [blank]*4]
        # Draw table
        the_table = plt.table(cellText=table_vals,
                              cellColours=colors,
                              colWidths=[0.05] * 4,
                              rowLabels=row_labels,
                              colLabels=col_labels,
                              loc='center')
        the_table.auto_set_font_size(False)
        the_table.set_fontsize(10)
        the_table.scale(4, 4)
        plt.title(title)
        # Removing ticks and spines enables you to get the figure only with table
        plt.tick_params(axis='x', which='both', bottom=False, top=False, labelbottom=False)
        plt.tick_params(axis='y', which='both', right=False, left=False, labelleft=False)
        for pos in ['right', 'top', 'bottom', 'left']:
            plt.gca().spines[pos].set_visible(False)
        if save_fig_path is not None:
            plt.savefig(save_fig_path)
        else:
            plt.show()
        plt.close()
        return True


def binary_clf_curve_with_ids(y_true, y_score, pos_label=None, sample_weight=None, y_true_ids=None):
    """Calculate true and false positives per binary classification threshold.

    Parameters
    ----------
    y_true : array, shape = [n_samples]
        True targets of binary classification

    y_score : array, shape = [n_samples]
        Estimated probabilities or decision function

    pos_label : int or str, default=None
        The label of the positive class

    sample_weight : array-like of shape = [n_samples], optional
        Sample weights.

    Returns
    -------
    fps : array, shape = [n_thresholds]
        A count of false positives, at index i being the number of negative
        samples assigned a score >= thresholds[i]. The total number of
        negative samples is equal to fps[-1] (thus true negatives are given by
        fps[-1] - fps).

    tps : array, shape = [n_thresholds <= len(np.unique(y_score))]
        An increasing count of true positives, at index i being the number
        of positive samples assigned a score >= thresholds[i]. The total
        number of positive samples is equal to tps[-1] (thus false negatives
        are given by tps[-1] - tps).

    thresholds : array, shape = [n_thresholds]
        Decreasing score values.
    """
    check_consistent_length(y_true, y_score)
    y_true = column_or_1d(y_true)
    y_score = column_or_1d(y_score)
    assert_all_finite(y_true)
    assert_all_finite(y_score)

    if sample_weight is not None:
        sample_weight = column_or_1d(sample_weight)

    # ensure binary classification if pos_label is not specified
    classes = np.unique(y_true)
    if (pos_label is None and
            not (np.array_equal(classes, [0, 1]) or
                 np.array_equal(classes, [-1, 1]) or
                 np.array_equal(classes, [0]) or
                 np.array_equal(classes, [-1]) or
                 np.array_equal(classes, [1]))):
        raise ValueError("Data is not binary and pos_label is not specified")
    elif pos_label is None:
        pos_label = 1.

    # make y_true a boolean vector
    y_true = (y_true == pos_label)

    # sort scores and corresponding truth values
    desc_score_indices = np.argsort(y_score, kind="mergesort")[::-1]
    if y_true_ids is not None:
        y_true_ids = [y_true_ids[i] for i in desc_score_indices]
    y_score = y_score[desc_score_indices]
    y_true = y_true[desc_score_indices]
    if sample_weight is not None:
        weight = sample_weight[desc_score_indices]
    else:
        weight = 1.

    # y_score typically has many tied values. Here we extract
    # the indices associated with the distinct values. We also
    # concatenate a value for the end of the curve.
    distinct_value_indices = np.where(np.diff(y_score))[0]
    threshold_idxs = np.r_[distinct_value_indices, y_true.size - 1]

    # accumulate the true positives with decreasing threshold
    tps = stable_cumsum(y_true * weight)[threshold_idxs]
    if sample_weight is not None:
        fps = stable_cumsum(weight)[threshold_idxs] - tps
    else:
        fps = 1 + threshold_idxs - tps

    return fps, tps, y_score[threshold_idxs], y_true_ids



#!/usr/bin/env python
"""Aws s3 interaction helper functions"""
########################################################################
# File: aws.py
#  executable: aws.py
#
# Author: Andrew Bailey
# History: 03/27/19 Created
########################################################################

import boto3
import botocore
import os
import warnings


class AwsS3(object):
    """Class to deal with getting information from aws s3"""
    def __init__(self, aws_access_key_id=None, aws_secret_access_key=None):

        self.s3_resource = boto3.resource('s3',
                                          aws_access_key_id=aws_access_key_id,
                                          aws_secret_access_key=aws_secret_access_key)
        self.s3_client = boto3.client('s3',
                                      aws_access_key_id=aws_access_key_id,
                                      aws_secret_access_key=aws_secret_access_key)
        self.connected = self.setup()

    def setup(self):
        """Hide unnecessary warnings"""
        warnings.filterwarnings("ignore", category=ResourceWarning, message="unclosed.*<ssl.SSLSocket.*>")
        try:
            [x for x in self.s3_resource.buckets.all()]
        except Exception as e:
            return False
        return True

    def download_object(self, path, dest):
        """Download file from specified path
        :param path: path to file
        :param dest: path to directory or renamed file output
        """
        bucket, key = self.split_name(path)
        if os.path.isdir(dest):
            dest = os.path.join(dest, os.path.basename(key))
        assert os.path.exists(os.path.dirname(dest)), \
            "Destination directory does not exist: {}".format(os.path.dirname(dest))
        try:
            self.s3_resource.Bucket(bucket).download_file(key, dest)
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == "404":
                print("The object does not exist. {}".format(path))
                dest = False
            else:
                raise e
        return dest

    def upload_object(self, file_path, destination, use_original_name=True):
        """Upload a file to s3 bucket
        :param use_original_name: boolean option to use the basename of original file
        :param file_path: path to file to upload
        :param destination: location to place file
        :return: True
        """
        assert os.path.exists(file_path), "File path does not exist {}".format(file_path)
        bucket_name, save_path = self.split_name(destination)
        if use_original_name:
            save_path = os.path.join(save_path, os.path.basename(file_path))

        self.s3_client.upload_file(file_path, bucket_name, save_path)
        return os.path.join(bucket_name, save_path)

    def create_bucket(self, bucket_name, region="us-west-2"):
        """Create a bucket
        :param bucket_name: name of bucket
        :param region: region to place bucket
        """
        self.s3_resource.create_bucket(Bucket=bucket_name, CreateBucketConfiguration={
            'LocationConstraint': region})
        return True

    def delete_bucket(self, bucket_name):
        """Delete a bucket
        :param bucket_name: name of bucket to delete
        :return: True
        """
        bucket = self.s3_resource.Bucket(bucket_name)
        for key in bucket.objects.all():
            key.delete()
        bucket.delete()
        return True

    def bucket_exists(self, bucket_name):
        """Check if bucket exists
        :param bucket_name: name of bucket to create
        :return: True if exists False if not
        """
        self.s3_resource.Bucket(bucket_name)
        exists = True
        try:
            self.s3_resource.meta.client.head_bucket(Bucket=bucket_name)
        except botocore.exceptions.ClientError as e:
            # If a client error is thrown, then check that it was a 404 error.
            # If it was a 404 error, then the bucket does not exist.
            error_code = e.response['Error']['Code']
            if error_code == '404':
                exists = False
        return exists

    def object_exists(self, object_path):
        """Return True if object exists
        :param object_path: path to object
        """
        bucket_name, save_path = self.split_name(object_path)
        try:
            self.s3_resource.Object(bucket_name, save_path).load()
        except botocore.exceptions.ClientError as e:
            return False
        return True

    def delete_object(self, object_path):
        """Delete object in s3
        :param object_path: path to object to delete
        """
        bucket_name, save_path = self.split_name(object_path)
        self.s3_client.delete_object(Bucket=bucket_name, Key=save_path)
        return True

    @staticmethod
    def split_name(name):
        """Split a name to get bucket and key path"""
        split_name = [x for x in name.split("/") if x is not '']
        bucket_name = split_name[0]
        key_path = "/".join(split_name[1:])
        return bucket_name, key_path

    def list_objects(self, path):
        """Return a list of full paths to objects within key or bucket
        :param path: path to bucket or key
        """
        bucket, key = self.split_name(path)
        my_bucket = self.s3_resource.Bucket(bucket)

        files = [os.path.join(bucket, file.key) for file in my_bucket.objects.filter(Prefix=key)
                 if self.object_exists(os.path.join(bucket, file.key))]
        return files

    def folder_exists(self, path):
        """Check to see if folder exists
        :param path: path to folder
        """
        bucket_name, save_path = self.split_name(path)
        try:
            result = self.s3_client.list_objects(Bucket=bucket_name, Prefix=save_path)
        except botocore.exceptions.ClientError as e:
            # The object does not exist.
            return False
        return True



#!/usr/bin/env python
"""Create setup script for pip installation of python_utils"""
########################################################################
# File: setup.py
#  executable: setup.py
#
# Author: Andrew Bailey
# History: 12/09/17 Created
########################################################################

import sys
from timeit import default_timer as timer
from setuptools import setup, find_packages


def main():
    """Main docstring"""
    start = timer()
    extras = {
        'seq_tools': ['pysam>=0.15', 'biopython>=1.68', 'mappy>=2.14']
    }
    setup(
        name="py3helpers",
        version='0.2.8',
        description='Python utility functions',
        url='https://github.com/adbailey4/python_utils',
        author='Andrew Bailey',
        license='MIT',
        setup_requires=["pytest-runner"],
        tests_require=["pytest"],
        author_email='andbaile@ucsc.com',
        packages=find_packages(),
        extras_requre=extras,
        install_requires=['numpy>=1.14.2',
                          'pandas>=0.23.4',
                          'scikit-learn>=0.19.0',
                          'matplotlib>=2.0.2',
                          'pytest-runner>=2.0',
                          'boto3>=1.9',
                          'pytest>=4.3.1'],
        zip_safe=True
    )

    stop = timer()
    print("Running Time = {} seconds".format(stop-start), file=sys.stderr)


if __name__ == "__main__":
    main()
    raise SystemExit



#!/usr/bin/env python
"""Utility functions for fasta, fastq and mapping reads"""
########################################################################
# File: seq_tools.py
#  executable: seq_tools.py
#
# Author: Andrew Bailey
# History: 02/15/18 Created
########################################################################

import sys
import numpy as np
import os
import mappy as mp
import re
import tempfile
from collections import defaultdict, namedtuple

from Bio.Seq import Seq
from Bio import pairwise2, SeqIO
from Bio.pairwise2 import format_alignment
from pysam import FastaFile, AlignmentFile, AlignedSegment, AlignmentHeader
from py3helpers.utils import merge_two_dicts, split_every_string

IUPAC_BASES = ("A", "C", "T", "G", "W", "R", "Y", "S", "K", "M", "B", "D", "H", "V", "N")
IUPAC_DICT = {"A": "A",
              "C": "C",
              "T": "T",
              "G": "G",
              "W": ["A", "T"],
              "R": ["A", "G"],
              "Y": ["C", "T"],
              "S": ["G", "C"],
              "K": ["G", "T"],
              "M": ["A", "C"],
              "B": ["C", "G", "T"],
              "D": ["A", "G", "T"],
              "H": ["A", "C", "T"],
              "V": ["A", "C", "G"],
              "N": "N"}

IUPAC_COMPLEMENTS = {"A": "T",
                     "C": "G",
                     "T": "A",
                     "G": "C",
                     "W": "W",
                     "R": "Y",
                     "Y": "R",
                     "S": "S",
                     "K": "M",
                     "M": "K",
                     "B": "V",
                     "D": "H",
                     "H": "D",
                     "V": "B",
                     "N": "N"}


def read_fasta(fasta, upper=True):
    """
    Taken from David Bernick but modified slightly to fit my purposes.

    using filename given in init, returns each included FastA record
    as 2 strings - header and sequence.
    whitespace is removed, no adjustment is made to sequence contents.
    The initial '>' is removed from the header.


    :param fasta: path to fasta file
    :param upper: boolean option to convert strings to uppercase
    :return: generator yielding header, sequence
    """
    assert fasta.endswith(".fa") or fasta.endswith(".fasta"), "Did not receive fasta file: {}".format(fasta)

    # initialize return containers

    def upperstring(line):
        return ''.join(line.rstrip().split()).upper()

    def lowerstring(line):
        return ''.join(line.rstrip().split())

    if upper:
        funct = upperstring
    else:
        funct = lowerstring

    sequence = ''
    with open(fasta, 'r+') as fasta_f:
        # skip to first fasta header
        line = fasta_f.readline()
        while not line.startswith('>'):
            line = fasta_f.readline()
        header = line[1:].rstrip()

        # header is saved, get the rest of the sequence
        # up until the next header is found
        # then yield the results and wait for the next call.
        # next call will resume at the yield point
        # which is where we have the next header
        for line in fasta_f:
            if line.startswith('>'):
                yield header, sequence
                # headerList.append(header)
                # sequenceList.append(sequence)
                header = line[1:].rstrip()
                sequence = ''
            else:
                sequence += funct(line)
        # final header and sequence will be seen with an end of file
        # with clause will terminate, so we do the final yield of the data

    yield header, sequence


def get_minimap_cigar(genome, sequence, preset='map-ont', cigar_string=True):
    """Get the alignment between a genome and alignment file

    :param genome: fasta file to genome
    :param sequence: sequence to align
    :param preset: sr for single-end short reads;
                   map-pb for PacBio read-to-reference mapping;
                   map-ont for Oxford Nanopore read mapping;
                   splice for long-read spliced alignment;
                   asm5 for assembly-to-assembly alignment;
                   asm10 for full genome alignment of closely related species.
    :param cigar_string: if True return normal cigar string, if false return array of shape (n_cigar, 2)
                        The two numbers give the length and the operator of each CIGAR operation.

    """
    assert os.path.exists(genome), "Genome path does not exist: {}".format(genome)
    assert preset in ["sr", "map-pb", "map-ont", "splice", "asm5", "asm10"]
    assert len(sequence) > 60, "minimap does not find alignments for small reads"
    a = mp.Aligner(genome, preset=preset)  # load or build index
    if not a:
        raise Exception("ERROR: failed to load/build index")
    for hit in a.map(sequence):
        if hit.is_primary:
            print(hit)
            if cigar_string:
                return str(hit.cigar_str)
            else:
                return hit.cigar


def get_minimap_alignment(genome, sequence, preset='map-ont'):
    """Get the alignment between a genome and alignment file

    :param genome: fasta file to genome
    :param sequence: sequence to align
    :param preset: sr for single-end short reads;
                   map-pb for PacBio read-to-reference mapping;
                   map-ont for Oxford Nanopore read mapping;
                   splice for long-read spliced alignment;
                   asm5 for assembly-to-assembly alignment;
                   asm10 for full genome alignment of closely related species.
    """
    assert os.path.exists(genome), "Genome path does not exist: {}".format(genome)
    assert preset in ["sr", "map-pb", "map-ont", "splice", "asm5", "asm10"]
    assert len(sequence) > 60, "minimap does not find alignments for small reads"
    a = mp.Aligner(genome, preset=preset)  # load or build index
    if not a:
        raise Exception("ERROR: failed to load/build index")
    for hit in a.map(sequence):
        if hit.is_primary:
            return hit


def create_pairwise_alignment(ref=None, query=None, local=False):
    """Generate alignment from two strings
    :param ref: string for reference squence
    :param query: string for query sequence
    :param local: if true do local alignment else do global
    :return: dictionary with 'reference' and 'query' accessible alignment sequences
    """
    assert ref is not None, "Must set reference sequence"
    assert query is not None, "Must set query sequence"
    if local:
        alignments = pairwise2.align.localms(ref.upper(), query.upper(), 2, -0.5, -1, -0.3,
                                             one_alignment_only=True)

    else:
        alignments = pairwise2.align.globalms(ref.upper(), query.upper(), 2, -0.5, -1, -0.3,
                                              one_alignment_only=True)
    # print(format_alignment(*alignments[0]))
    return {'reference': alignments[0][0], 'query': alignments[0][1]}


def get_pairwise_cigar(ref=None, query=None, local=False):
    """Generate alignment from two strings
    :param ref: string for reference squence
    :param query: string for query sequence
    :param local: bool for local or global alignment
    """
    assert ref is not None, "Must set reference sequence"
    assert query is not None, "Must set query sequence"

    alignment = create_pairwise_alignment(ref=ref, query=query, local=local)
    final_str = str()
    # CIGAR = {"M": 0, "I": 0, "D": 0, "N": 0, "S": 0, "H": 0, "P": 0, "=": 0, "X": 0}
    current = str()
    count = 0
    for ref, query in zip(alignment['reference'], alignment["query"]):
        if ref == query:
            # matches
            if current == "M":
                count += 1
                current = "M"
            else:
                final_str += str(count) + current
                current = "M"
                count = 1
        elif ref == '-':
            # soft clipped sequences
            if current == "S":
                count += 1
                current = "S"
            elif current == str():
                final_str += str(count) + current
                current = "S"
                count = 1
            # insertions
            elif current == "I":
                count += 1
                current = "I"
            else:
                final_str += str(count) + current
                current = "I"
                count = 1
        elif query == '-':
            # deletions
            if current == "D":
                count += 1
                current = "D"
            else:
                final_str += str(count) + current
                current = "D"
                count = 1
        else:
            # mismatches
            if current == "X":
                count += 1
                current = "X"
            else:
                final_str += str(count) + current
                current = "X"
                count = 1
    if current == "I":
        final_str += str(count) + "S"
    else:
        final_str += str(count) + current
    # remove initial zero
    return final_str[1:]


def cigar_conversion(cigar):
    """Convert character or integer to corresponding integer or character for the SAM/BAM formatting
    :param cigar: character or int to get corresponding CIGAR equivalent
    """
    assert type(cigar) is str or type(cigar) is int, "Cigar character must be int or str"
    chars = ["M", "I", "D", "N", "S", "H", "P", "=", "X"]

    if type(cigar) is str:
        assert cigar in chars, "Character is not a cigar string character"
        return chars.index(cigar)
    else:
        assert 0 <= cigar <= 8, "Integer must be between 0 and 8"
        return chars[cigar]


class ReverseComplement(object):
    """Class to deal with reverse, complement and reverse-complement sequences"""

    def __init__(self, find="ATGC", replace="TACG"):
        """Set find and replace characters. Entire class will turn everything into upper case

        :param find: strings to find then replace with replace
        :param replace: string which replaces the corresponding index of parameter find
        """
        assert len(find) == len(set(find)), "No repeats allowed in param 'find': {}".format(find)
        assert len(replace) == len(set(replace)), "No repeats allowed in param 'replace': {}".format(replace)
        assert len(find) == len(replace), "Length of 'find' must match 'replace': len({}) != len({})".format(find,
                                                                                                             replace)
        self.find = find.upper()
        self.replace = replace.upper()
        self.transtab = str.maketrans(self.find, self.replace)

    def complement(self, string):
        """Generate complement sequence"""
        return string.upper().translate(self.transtab)

    @staticmethod
    def reverse(string):
        """Reverse string sequence"""
        return string.upper()[::-1]

    def reverse_complement(self, string):
        """Reverse complement string from transtable"""
        string = self.reverse(string)
        return self.complement(string)

    def convert_write_fasta(self, in_fasta, outpath, complement=True, reverse=True):
        """Write new fasta file to outpath while adding info to header names

        :param reverse: boolean option for reverse direction of string
        :param complement: boolean option for complement of string
        :param in_fasta: input fasta file
        :param outpath: path to output fasta """
        assert outpath.endswith(".fa") or outpath.endswith(".fasta"), "Output file needs to end with .fa or .fasta"
        assert os.path.isfile(in_fasta), "If fasta is selected, sequence must be a real path"
        assert complement or reverse is True, "Must select complement or reverse"

        # choose function and information to add to header
        if complement and not reverse:
            funct = self.complement
            add_to_header = "_complement"
        elif reverse and not complement:
            funct = self.reverse
            add_to_header = "_reverse"
        elif complement and reverse:
            funct = self.reverse_complement
            add_to_header = "_reverse_complement"
        # write to fasta file
        with open(outpath, 'w+') as out_fa:
            for header, sequence in read_fasta(in_fasta):
                out_fa.write('>' + header + add_to_header + '\n')
                for sub_seq in split_every_string(80, funct(sequence)):
                    out_fa.write(sub_seq + '\n')
        return outpath

    def convert_write_fastq(self, in_fastq, outpath, complement=True, reverse=True):
        """Write new fastq file to outpath while adding info to header names

        :param reverse: boolean option for reverse direction of string
        :param complement: boolean option for complement of string
        :param in_fastq: input fastq file
        :param outpath: path to output fastq """
        assert outpath.endswith(".fq") or outpath.endswith(".fastq"), \
            "Output file needs to end with .fq or .fastq: {}".format(outpath)
        assert os.path.isfile(in_fastq), "If fastq is selected, sequence must be a real path: {}".format(in_fastq)
        assert complement or reverse is True, "Must select complement and/or reverse"

        # choose function and information to add to header
        if complement and not reverse:
            funct = self.complement
            add_to_header = "_complement"
        elif reverse and not complement:
            funct = self.reverse
            add_to_header = "_reverse"
        elif complement and reverse:
            funct = self.reverse_complement
            add_to_header = "_reverse_complement"
        # write to fastq file
        reads = []
        for record in SeqIO.parse(in_fastq, "fastq"):
            record.id = record.id + add_to_header
            record.seq = Seq(funct(str(record.seq)))
            if reverse:
                record.letter_annotations["phred_quality"] = record.letter_annotations["phred_quality"][::-1]
            reads.append(record)
        SeqIO.write(reads, outpath, "fastq")
        return outpath


def write_fasta(header_sequence_dict, outpath):
    """Write header and sequence with correct fasta formatting. Sequence converted to uppercase

    :param header_sequence_dict: dict of header with sequence
    :param outpath: path to file ending in .fa or .fasta
    """
    assert outpath.endswith(".fa") or outpath.endswith(".fasta"), "Output file needs to end with .fa or .fasta"
    assert type(header_sequence_dict) is dict, "header_sequence_dict needs to be a dictionary"

    with open(outpath, 'w+') as out_fa:
        for header, sequence in header_sequence_dict.items():
            out_fa.write('>' + header + '\n')
            for subset in split_every_string(80, sequence.upper()):
                out_fa.write(subset + '\n')


def check_fastq_line(fastq):
    """Check if fastq string has some necessary features of the fasta format.

    :param fastq: single string with a record of fastq """

    check = fastq.split('\n')
    assert len(check) == 4, "Data is not fastq format: Not enough fields"
    assert check[0].startswith('@'), "Data is not fastq format: Does not start with @"
    assert len(check[1]) == len(check[3]), "Data is not fastq format: Sequence and quality scores do not match"
    assert check[2].startswith('+'), "Data is not fastq format: third line does not start with +"
    return True


def create_fastq_line(read_id, sequence, q_values):
    """Create a fastq string from the necessary fields. Do not include newlines!

    :param read_id: unique identifier for read
    :param sequence: sequence of nucleotides
    :param q_values: quality score values associated with the sequence
    """
    # make sure we have not included newline characters
    assert read_id.find("\n") == -1, "Remove newline characterss from read_id"
    assert sequence.find("\n") == -1, "Remove newline characters from sequence"
    assert q_values.find("\n") == -1, "Remove newline characters from q_values"
    assert len(sequence) == len(q_values), "sequence and q_values must to be the same size"

    if not read_id.startswith("@"):
        read_id = '@' + read_id
    line1 = read_id + "\n"
    line2 = sequence + "\n"
    line3 = "+\n"
    fastq = line1 + line2 + line3 + q_values
    return fastq


class SeqAlignment(object):
    """Keep track of alignment accuracy statistics from two sequences and a cigar string"""
    CIGAR = ["M", "I", "D", "N", "S", "H", "P", "=", "X"]

    # S= soft clip (ignore)
    # H = Hard clip (ignore)
    # N = ignore length for splice assignments (ignore)
    # P = padded reference (ignore)
    # M = Match
    # I = Insertion
    # D = Deletion
    # = = match
    # X = Mismatch

    def __init__(self, query, ref, cigar):
        """Initialize summary accuracy information from a cigar alignment

        :param query: query sequence (can be soft padded)
        :param ref: reference sequence (only aligned part)
        :param cigar: cigar string
        """
        self.query = query.upper()
        self.ref = ref.upper()
        self.cigar = cigar
        # contains alignment for each base in the query and reference positions
        self.query_map = []
        self.ref_map = []
        # contains the entire alignment and true matches mapping
        # full_alignment contains soft clipped sections
        self.full_alignment_map = []
        # alignment map does not contain clipped regions
        self.alignment_map = []
        self.true_matches_map = []
        self.matches_map = []
        # each key h
        self.base_alignment = namedtuple('base_alignment',
                                         ['query_index', 'query_base', 'reference_index', 'reference_base'])
        self.alphabet = set(self.ref + self.query)
        # self.totals = namedtuple('totals', ['matches', 'inserts', 'deletes', 'clipped', 'mismatch])

        self.counts = {char: [0, 0, 0, 0, 0] for char in self.alphabet}
        assert self._initialize()

    def _initialize(self):
        """Get percent of identically matched bases over the alignment length"""
        cigar_chars = set(''.join([i for i in self.cigar if not i.isdigit()]))
        assert cigar_chars.issubset(self.CIGAR), \
            "Cigar contains extraneous characters. Only 'MIDNSHP=X' allowed. {}".format(cigar_chars)
        # track indexes
        query_index = 0
        ref_index = 0
        total_len = 0
        # go through cigar
        for num, cigar_char in re.findall('(\d+)([MIDNSHP=X])', self.cigar):
            num = int(num)
            total_len += num
            # map bases to reference
            for x in range(num):
                if cigar_char == "M" or cigar_char == "=" or cigar_char == "X":
                    # matches and mismatches
                    align = self.base_alignment(query_index, self.query[query_index],
                                                ref_index, self.ref[ref_index])
                    # consumes both
                    self.query_map.append(align)
                    self.ref_map.append(align)
                    self.alignment_map.append(align)
                    self.full_alignment_map.append(align)
                    self.matches_map.append(align)
                    if self.ref[ref_index] == self.query[query_index]:
                        self.true_matches_map.append(align)
                        self.counts[self.query[query_index]][0] += 1
                    else:
                        self.counts[self.query[query_index]][4] += 1

                    query_index += 1
                    ref_index += 1
                elif cigar_char == "S" or cigar_char == "I":
                    # skips and inserts
                    align = self.base_alignment(query_index, self.query[query_index],
                                                None, None, )
                    # consumes query
                    self.query_map.append(align)
                    if cigar_char != "S":
                        self.alignment_map.append(align)
                        self.counts[self.query[query_index]][3] += 1
                    else:
                        self.counts[self.query[query_index]][1] += 1

                    self.full_alignment_map.append(align)
                    query_index += 1
                elif cigar_char == "D" or cigar_char == "N":
                    # skipped regions for splice assignments so not used for alignment length but is used for ref length
                    align = self.base_alignment(None, None,
                                                ref_index, self.ref[ref_index])
                    # consumes reference
                    self.ref_map.append(align)
                    self.alignment_map.append(align)
                    self.full_alignment_map.append(align)
                    if cigar_char == "D":
                        self.counts[self.ref[ref_index]][2] += 1
                    ref_index += 1
                elif cigar_char == "P":
                    # currently passing
                    pass

        assert len(self.full_alignment_map) == total_len, "Full alignment map does not match cigar length. Check inputs"
        assert len(self.ref_map) == len(self.ref), "Ref map does not match ref seq. map: {} != ref_len: {}" \
            .format(len(self.ref_map), len(self.ref))
        assert len(self.query_map) == len(self.query), "Query map does not match query seq. map: {} != query_len: {}" \
            .format(len(self.query_map), len(self.query))

        return True

    def alignment_accuracy(self, soft_clip=False):
        """Get percent accuracy from cigar string

        :param soft_clip: boolean include soft clip in accuracy calculation
        """
        if soft_clip:
            acc = float(len(self.true_matches_map)) / len(self.full_alignment_map)
        else:
            acc = float(len(self.true_matches_map)) / len(self.alignment_map)

        return acc


class Cigar(object):
    """Cigar string object to handle cigar strings"""
    CIGAR = ["M", "I", "D", "N", "S", "H", "P", "=", "X"]

    # S= soft clip (ignore)
    # H = Hard clip (ignore)
    # N = ignore length for splice assignments (ignore)
    # P = padded reference (ignore)
    # M = Match
    # I = Insertion
    # D = Deletion
    # = = match
    # X = Mismatch

    def __init__(self, cigar):
        """Get some basic info from a cigar string

        :param cigar: cigar string
        """
        self.cigar = cigar
        self.ref_len = 0
        self.query_len = 0
        self.alignment_length = 0
        self.alignment_length_soft_clipped = 0
        self.matches = 0
        self.indexed_cigar = []
        self._initialize()

    def _initialize(self):
        """Initialize important information from cigar string"""
        cigar_chars = set(''.join([i for i in self.cigar if not i.isdigit()]))
        assert cigar_chars.issubset(
            self.CIGAR), "Cigar contains extraneous characters. Only 'MIDNSHP=X' allowed. {}".format(cigar_chars)
        soft_clipped = 0
        for num, char in re.findall('(\d+)([MIDNSHP=X])', self.cigar):
            num = int(num)
            if char == "M" or char == "=":
                self.matches += num
                self.alignment_length += num
                self.ref_len += num
                self.query_len += num
                self.indexed_cigar.extend(["M" for _ in range(num)])
            elif char == "X":
                self.alignment_length += num
                self.ref_len += num
                self.query_len += num
                self.indexed_cigar.extend(["X" for _ in range(num)])
            elif char == "S":
                self.query_len += num
                soft_clipped += num
                self.indexed_cigar.extend(["S" for _ in range(num)])
            elif char == "I":
                self.query_len += num
                self.alignment_length += num
                self.indexed_cigar.extend(["I" for _ in range(num)])
            elif char == "D":
                self.ref_len += num
                self.alignment_length += num
                self.indexed_cigar.extend(["D" for _ in range(num)])
            elif char == "N":
                # skipped regions for splice assignments so not used for alignment length but is used for ref length
                self.ref_len += num
                self.indexed_cigar.extend(["N" for _ in range(num)])
            elif char == "P":
                self.indexed_cigar.extend(["P" for _ in range(num)])
                # do nothing for padded regions
                pass

        self.alignment_length_soft_clipped = self.alignment_length + soft_clipped

    def accuracy_from_cigar(self, soft_clip=False):
        """Get percent accuracy from cigar string

        :param soft_clip: boolean include soft clip in accuracy calculation
        """
        if soft_clip:
            acc = float(self.matches) / self.alignment_length_soft_clipped
        else:
            acc = float(self.matches) / self.alignment_length

        return acc

    def reverse_cigar(self):
        """Reverse cigar sequence"""
        reversed_cigar = self.indexed_cigar[::-1]
        new_cigar = ""
        prev_char = reversed_cigar[0]
        counter = 1
        for x in reversed_cigar[1:]:
            if x == prev_char:
                counter += 1
            else:
                new_cigar += str(counter) + prev_char
                prev_char = x
                counter = 1
        new_cigar += str(counter) + prev_char
        return new_cigar


def pairwise_alignment_accuracy(ref=None, query=None, local=False, soft_clip=False):
    """Get accuracy from pairwise alignment of two sequences

    :param ref: string 1 to treat as reference
    :param query: string 2 to treat as query
    :param local: boolean for local or global alignment
    :param soft_clip: boolean option for including soft_clipped regions for accuracy
    """
    # complete pairwise alignment and generate cigar string
    assert ref is not None, "Must set reference sequence"
    assert query is not None, "Must set query sequence"
    cigar = get_pairwise_cigar(ref, query, local=local)
    accuracy = Cigar(cigar).accuracy_from_cigar(soft_clip=soft_clip)
    return accuracy


class ReferenceHandler:
    """
    This class handles fasta reference files, ensuring that the sequence is not a terminal 'N' and that the end of the
    sequence has not been reached
    """

    def __init__(self, reference_file_path):
        """
        create fasta file object given file path to a fasta reference file
        :param fasta_file_path: full path to a fasta reference file
        """

        self.fasta_file_path = reference_file_path
        assert os.path.exists(reference_file_path), "Reference path does not exist: {}".format(reference_file_path)
        try:
            self.fasta = FastaFile(self.fasta_file_path)
        except Exception as e:
            print(e)
            raise IOError("Fasta File Read Error: Try indexing reference with 'samtools faidx {}'"
                          .format(reference_file_path))

    def get_sequence(self, chromosome_name, start, stop):
        """
        Return the sequence of a chromosome region by zero based indexing: chr[start:stop]

        :param chromosome_name: Chromosome name
        :param start: Region start index
        :param stop: Region end, one more than last index
        :return: Sequence of the region
        """
        return self.fasta.fetch(reference=chromosome_name, start=start, end=stop).upper()

    def get_chr_sequence_length(self, chromosome_name):
        """
        Get sequence length of a chromosome. This is used for selecting windows of parallel processing.
        :param chromosome_name: Chromosome name
        :return: Length of the chromosome reference sequence
        """
        return self.fasta.get_reference_length(chromosome_name)

    def write_new_reference(self, outpath, reverse=False, complement=False, find="ATGC", replace="TACG"):
        """Write a new reference sequence file to a given directory with options of reverse and/or complement the strand

        ex... directory = '/home/name/dir/', reverse=True, complement=True
              return = /home/name/dir/reference.reverse.complement.fa

        :param outpath: path to directory
        :param reverse: bool option to reverse sequence
        :param complement: bool option to complement sequence
        :param find: strings to find then replace with replace
        :param replace: string which replaces the corresponding index of parameter find
        :return: path to file if pass otherwise False
        """
        # get original fasta name
        fasta_name = os.path.splitext(os.path.basename(self.fasta_file_path))[0]
        # edit fasta name
        if reverse:
            fasta_name += '.reverse'
        if complement:
            fasta_name += '.complement'
        fasta_name += '.fa'
        # create outpath
        outpath = os.path.join(outpath, fasta_name)
        # write fasta with edits
        ReverseComplement(find=find, replace=replace).convert_write_fasta(self.fasta_file_path, outpath,
                                                                          complement=complement, reverse=reverse)
        return outpath


def initialize_pysam_wrapper(sam_file_string, reference_path=None):
    """Create a temp file to initialize the pysam wrapper class
    :param sam_file_string: correctly formatted SAM alignment string
    :param reference_path: path to reference file
    """
    with tempfile.TemporaryDirectory() as tempdir:
        path = os.path.join(tempdir, "test.sam")
        with open(path, "w") as tmp:
            tmp.write(sam_file_string)
        try:
            pysam_handle = PysamWrapper(path)
            pysam_handle.initialize(reference_path)
            return pysam_handle
        except StopIteration:
            raise ValueError("sam_file_string is not in SAM format")


class PysamWrapper(AlignmentFile):
    """Use pysam AlignmentFile functions for a string sam entry"""

    def __init__(self, path):
        """Create wrapper for pysam to deal with single alignment sam files

        :param path: path to temp single alignment sam file

        """
        try:
            self.alignment_segment = self.fetch().__next__()
        except OSError:
            raise OSError("Check SAM cigar/ MDZ field formatting. Truncated File error")
        self.reference_path = None
        self.seq_alignment = None
        self.initialized = False

    def initialize(self, reference_path=None):
        """Initialization function for Pysam Wrapper class

        Since AlignmentFile is in Cython, could not get the inheritance to work correctly

        :param reference_path: path to reference genome
        """
        self.reference_path = reference_path
        self.seq_alignment = SeqAlignment(cigar=self.get_cigar(),
                                          ref=self.get_reference_sequence(),
                                          query=self.alignment_segment.get_forward_sequence())
        self.initialized = True

    def get_reference_sequence(self):
        """Get reference sequence from sequence position. Will reverse if alignment is to reverse strand"""
        if not self.alignment_segment.has_tag('MD'):
            assert self.reference_path is not None, "Need to designate reference path if MD:Z field is not defined"
            fasta_handle = ReferenceHandler(self.reference_path)
            ref_sequence = fasta_handle.get_sequence(chromosome_name=self.alignment_segment.reference_name,
                                                     start=self.alignment_segment.reference_start,
                                                     stop=self.alignment_segment.reference_end)
        else:
            ref_sequence = self.alignment_segment.get_reference_sequence().upper()

        if self.alignment_segment.is_reverse:
            ref_sequence = ReverseComplement().reverse_complement(ref_sequence)

        return ref_sequence

    def get_cigar(self):
        """Get cigar string. Will reverse if alignment is to reverse strand"""
        cigar = self.alignment_segment.cigarstring
        # deal with reversed alignment
        if self.alignment_segment.is_reverse:
            cigar = Cigar(cigar).reverse_cigar()
        return cigar

    def alignment_accuracy(self, soft_clip=False):
        """Get alignment accuracy
        :param soft_clip: boolean include soft clip in accuracy calculation
        """
        return self.seq_alignment.alignment_accuracy(soft_clip=soft_clip)


class AlignmentSegmentWrapper(object):
    """Use pysam AlignmentSegment functions for a string sam entry"""

    def __init__(self, aligned_segment):
        """Create wrapper for pysam to deal with single alignment sam files

        :param aligned_segment: AlignmentSegment object

        """
        self.alignment_segment = aligned_segment
        self.reference_path = None
        self.seq_alignment = None
        self.initialized = False

    def initialize(self, reference_path=None):
        """Initialization function for Pysam Wrapper class

        Since AlignmentFile is in Cython, could not get the inheritance to work correctly

        :param reference_path: path to reference genome
        """
        self.reference_path = reference_path
        self.seq_alignment = SeqAlignment(cigar=self.get_cigar(),
                                          ref=self.get_reference_sequence(),
                                          query=self.alignment_segment.get_forward_sequence())
        self.initialized = True

    def get_reference_sequence(self):
        """Get reference sequence from sequence position. Will reverse if alignment is to reverse strand"""
        if not self.alignment_segment.has_tag('MD'):
            assert self.reference_path is not None, "Need to designate reference path if MD:Z field is not defined"
            fasta_handle = ReferenceHandler(self.reference_path)
            ref_sequence = fasta_handle.get_sequence(chromosome_name=self.alignment_segment.reference_name,
                                                     start=self.alignment_segment.reference_start,
                                                     stop=self.alignment_segment.reference_end)
        else:
            ref_sequence = self.alignment_segment.get_reference_sequence().upper()

        if self.alignment_segment.is_reverse:
            ref_sequence = ReverseComplement().reverse_complement(ref_sequence)

        return ref_sequence

    def get_cigar(self):
        """Get cigar string. Will reverse if alignment is to reverse strand"""
        cigar = self.alignment_segment.cigarstring
        # deal with reversed alignment
        if self.alignment_segment.is_reverse:
            cigar = Cigar(cigar).reverse_cigar()
        return cigar

    def alignment_accuracy(self, soft_clip=False):
        """Get alignment accuracy
        :param soft_clip: boolean include soft clip in accuracy calculation
        """
        return self.seq_alignment.alignment_accuracy(soft_clip=soft_clip)


def sam_string_to_aligned_segment(sam_string, header=None):
    """Convert a correctly formatted sam string into a pysam AlignedSegment object

    :param sam_string: correctly formatted SAM string
    :param header: AlignmentHeader object

    :return AlignedSegment
    """
    if not header:
        header = AlignmentHeader.from_references([sam_string.split("\t")[2]], [100000000])

    new_segment = AlignedSegment.fromstring(sam_string, header)

    return new_segment


def initialize_aligned_segment_wrapper(sam_string, reference_path=None):
    """Convert a correctly formatted sam string into a pysam AlignedSegment object

    :param sam_string: correctly formatted SAM string
    :param reference_path:

    :return AlignmentSegmentWrapper
    """

    segment = sam_string_to_aligned_segment(sam_string)
    as_wrapper = AlignmentSegmentWrapper(segment)
    as_wrapper.initialize(reference_path)
    return as_wrapper


def iupac_complement(nuc):
    """Return the complement character using IUPAC nucleotides"""
    nuc = nuc.upper()
    assert is_iupac_base(nuc), "Nucleotide is not IUPAC character"
    return IUPAC_COMPLEMENTS[nuc]


def iupac_base_to_bases(nuc):
    """Return the bases that are represented for a given iupac base"""
    nuc = nuc.upper()
    assert is_iupac_base(nuc), "Nucleotide is not IUPAC character"
    return IUPAC_DICT[nuc]


def is_iupac_base(nuc):
    """Returns True if an IUPAC base and False if not"""
    if nuc.upper() in IUPAC_BASES:
        return True
    else:
        return False


def is_non_canonical_iupac_base(nuc):
    """Return True if base is one of teh IUPAC bases but not ATGC"""
    nuc = nuc.upper()
    if nuc in IUPAC_BASES and nuc not in "ATGC":
        return True
    else:
        return False


if __name__ == '__main__':
    test_sam = "3d7384c3-845b-4d54-9ba4-75c7992382b9    16      Chromosome      3604742 60      25S20M2D4M1D84M1D13M3D4M1D44M1D21M1D62M3D5M1D6M1D4M1D16M1D6M1D6M1I15M2D19M1D7M5D9M1D30M1D30M1D27M2I35M2D11M1D21M2I14M1D17M2I32M4D49M1I37M1D3M1I16M1D8M1D32M1D9M2I10M1D1M1D14M1D14M1D49M1D25M1I28M1I9M1I24M1D14M4D27M1D11M2D17M1D23M1I100M3I23M1D9M1D8M2D44M2I28M10D44M1D3M1D12M1I22M4D5M1D17M1D80M3D48M1D19M1D34M1I17M3D47M1D32M1D25M1D6M1D24M1I22M1D16M2I8M3D61M3D4M1D6M1D35M1I20M3D6M2D29M1D1M5D19M2D5M1D12M1D2M3D28M1D2M2D16M1D6M2I64M5D3M1D20M1I6M2D47M1D12M1D11M2D32M3D5M1I10M3I34M1I28M1D1M2D5M1D9M1D1M7D1M3D20M1D6M3I7M2D24M1I11M1D10M1D83M2D12M2D1M1D30M1D6M1D6M7D13M1D56M2D12M1D47M1I11M4D7M1D3M1D6M1D9M5D5M1I3M1I5M1D30M1D8M1D15M1D6M3D33M3D1M2D3M1D6M1D13M2D10M2D5M2D7M1D9M2D17M1D13M2D5M1D5M1D2M2D5M1D13M1I8M3D3M2D20M1I6M2D3M1D58M1D39M1D24M2D7M1I10M1D3M3D7M3I12M2I23M3D50M1D3M1D31M2D2M1D23M6D7M1I30M4D6M2D9M1D39M1D6M3D41M2D25M2D21M1I5M1D34M2D5M1D65M2I12M1D27M1I40M2D9M1D28M3D7M1D8M1D9M1I6M1D55M1I6M1I24M1D1M1D19M1D3M2D19M3D22M1D47M1I29M1I12M1D4M3D32M1I21M2I28M2D23M2D9M1D6M1D16M3D12M3D30M1I2M1D4M1D19M1D9M2I42M2D52M3D22M1D9M1D105M1D5M2D19M1D15M2D6M1D5M2D22M2D19M2I19M1D12M1D2M1D4M1D24M2D89M1D9M1D50M1D100M1D18M2I9M1D19M1D7M6D6M1D12M1D30M2D13M1D52M1D24M1I36M1D21M1D13M1D40M2D42M2D3M3D1M1D23M2I9M1D17M1D45M1D31M1D4M6D10M1D33M1D11M2I9M1D19M1D27M1I11M2I13M1D12M2D5M2D31M1D15M1D1M1D17M6D6M4D18M2D1M1D23M1D11M1D5M3D16M1I2M1D14M1I14M1I11M3I16M1I7M1I50M1D9M1D12M5D13M1I10M1D11M2D29M4D10M1D10M1D25M1I48M1D2M1D4M3I24M5D19M1I14M1D16M1I16M2I2M2I46M2D54M1I32M1D15M1D3M1D7M1D25M1D4M2D3M1D22M2D3M1D41M1D10M1D37M1D20M1I18M1D7M1I10M1D87M1D30M3D17M1D11M1D50M1D18M1I31M3D31M2I9M3D4M1D13M2D5M3D30M4D5M1I18M1I19M1D1M1D13M1I7M1D9M1I21M2D2M1D2M1D13M2D2M1D5M1D1M1D20M1D13M1D4M1D4M2I5M2I24M3D10M1D21M1I5M2D10M2I11M2D12M1D41M2D132M1D53M1I23M2D15M3D9M1D6M1D5M2D1M1D12M1D45M2D11M1I8M1I36M1D3M1D15M1I36M1I23M1D23M1I5M1I14M2D34M4I34M1D5M1I1M1I94M4D3M1D11M2D24M1D13M26S *       0       0         *       NM:i:1054       MD:Z:2G17^GT4^G0T38T8T21T13^T0T0T0A0T9^GAA4^A15T12T2T2T0T0A7^A5T15^T0T0A56T0T2^TAA5^T6^G4^T16^G1T4^C6C0A7T5^TC4T14^T7^GTGCT9^T0A0T28^C12C5T11^A15C1T10T0T0T31^TT0A10^T35^T0T0T0A26T19^GGGC86^C4T0G13^T8^G8C8C1C12^T19^A1^C14^C3T8C1^T0A13C34^A47T10T27^C12A1^TGGA27^A5T5^AT0T16^T18C0T101T2C0A0A19^T0T2C5^G1T6^AG15T14T18A3C0T17^CGTTTTTCCT8T0G1T19A6T5^T3^T31T0T1^TGGG5^T1T15^G21T4C0T25T10T15^TGT30C0T16^C1T13T3^T51^TTG7C0T1C34T1^T0A0T30^C25^A6^T26T19^C12T3T0T6^GCT61^CTC4^A6^C9C2T42^GCA6^AA29^T1^AGGTG9T9^AA5^G1T2T7^G2^AAT28^T2^AA16^T68T1^ATCGT3^G1T22C1^GT0T46^G3T5C2^T0T1C8^CA32^CAT6C10T17T13A22G0C0T2^T1^TT0T0T0T2^T0T5A2^A1^TGGAAGG1^TGG20^G6A6^TA1T33^T0A0A0G7^G8T14T59^TG1T10^GC1^A30^C6^G6^CGCGCAA13^C40A5T9^GG12^T47C1A0T2A0G3^TTTA7^T0T2^A4T1^T9^CGCGA1T7T3^T13T16^T0T0T6^A2T12^T0T0T1T0A1^AGG2T5T13T7T2^GTT1^GT3^C6^C2G0C9^CT10^GG0C4^GC7^G9^TT0A16^G7T5^GT1T0T2^T5^G2^GT5^A21^GTA3^CG17T8^GG3^T0T57^C0T38^T24^GC17^G3^TTG0A41^GGA7C2T14T0T0T0A0A1T18^G3^A21T5T3^GT2^T0T20T1^CCCTGC16T20^GCTG6^AA2G6^C39^G6^TGG0G7T2A0A10T17^AA25^GC21C0G3^T0T33^GG5^C0A1T74^A51T3G0C3G6^AA0A0G7^C4T23^GGT1G0T4^G0G7^G9G2G2^A0C11C0A49C11T6C2^C1^C19^G3^GA2G16^AGC22^T34T0C11C0A39^G4^GTG32G0C7C2A1T0G0C32^GG1C7G0C12^GT9^G6^A16^TTC12^AAA32^T0A0A2^A3T7C7^T34T16^AA52^GGT0T21^A9^A1T103^T1T3^TT0T0T17^G15^TC6^T5^TT22^TT38^A1T10^T0T1^T0T1A1^C24^TA10T78^T0G8^C5C23T20^G76C23^G27^T0T18^T7^TTCCTG0C5^T0A0T10^T30^AT13^T52^A1C39A0T1T15^A1T8T4A0T4^C13^T0G39^AA0A1T28T0T9^TT0A2^TTT1^T3T19C0A7^A17^A1T6C4C0T4C19C2T2^T20C10^C4^GGTGCT10^C1G17C0C6C5^T2A0A7C8^C19^A11T0T0A37^C12^GG5^GT22C1A0A0T0T1C1^C15^A1^C0A1A0A2C10^GCAGAA6^ATTA1T9A6^AA1^T23^G11^G5^CGT1T12C3^A1T0A1T10C96^A2T1C4^T0A0A1T8^CCGCC14C1G3C2^T11^TA29^CCGG0C1C7^T4C2C2^T17T7A47^G0C1^A4T23^GTTAG1G2T10T0A16^T0A42T1A0T0A8T1A0T0A19^AT3G0C66T14^T15^G3^T7^A7C17^A1T2^TT0A2^G1T20^GT3^A17G0C5C13C2^A0A9^C17C3C15^A15C0A1A9C9^A15C1^A1G85^A0A0A28^TCT17^T11^C50^A1T16T5C9A0T10C2^TAA12C4C13C0T7^TTA4^T13^AT3C1^AAT0T11C17^CTAC1G4G0G21T0A11^C1^T5C7T6^A30^GC2^G2^T13^AT2^A5^T1^A20^C13^A4^G4T28^GTA10^C0G22T2^TT0G0T8T10^TT12^C7C33^AC72C26T32^T54A1T2T16^GA5T9^CAC9^A1T4^G5^TC1^G12^G0C44^GC9C2A0C0A0A0A0A12T24^G3^T0T4C1C7A1G8C21T10T14^G0T1T7G1T5T4G0G0G3C9T2^TG68^A1T32A1T34T28^GGAT3^G11^GG4G0T18^T13        AS:i:5652       XS:i:0"
    reference = "/Users/andrewbailey/CLionProjects/nanopore-RNN/test_files/reference-sequences/ecoli_k12_mg1655.fa"
    # test_write_new_reference()
    # sam_h = SamHandler(test_sam)
    # # sam_h.check_sam_format()
    # print(sam_h.cigar_accuracy())
    # print(sam_h.get_reference_sequence(reference=reference))

    print("This is a library of sequence helper functions")
